__author__ = 'smackware'


try:
    from cPickle import loads, dumps
except ImportError:
    from pickle import loads, dumps


class AbstractSerializer(object):
    def dumps(self, obj):
        raise NotImplementedError("encode is not implemented")

    def loads(self, obj):
        raise NotImplementedError("decode is not implemented")


class PickleSerializer(object):
    @staticmethod
    def dumps(obj):
        return dumps(obj)

    @staticmethod
    def loads( data):
        return loads(data)


__author__ = 'smackware'

from .payload import RequestPayload

from .identification import generate_task_id
from .identification import get_request_key

from .response import RdisqResponse


class RdisqAsyncConsumer(object):
    def register_call_(self, name, call):
        setattr(self, name, call)


class AbstractRdisqConsumer(object):
    service_class = None

    def __init__(self, service_class):
        self.__queue_to_callable = None
        self.service_class = service_class
        self.__setup_stub_methods_for_consumer()

    def __setup_stub_methods_for_consumer(self):
        self.__queue_to_callable = {}
        for attr in dir(self.service_class):
            call = getattr(self.service_class, attr)
            if not self.service_class.is_remote_method(call):
                continue
            stub_method_name = self.service_class.chop_prefix_from_exported_method_name(attr)
            setattr(self, stub_method_name, self.get_stub_method(stub_method_name))
            self.__queue_to_callable[stub_method_name] = call
        if not self.__queue_to_callable:
            raise AttributeError("Cannot instantiate a consumer with no exposed methods")

    def send(self, method_name, *args, **kwargs):
        timeout = kwargs.pop("timeout", self.service_class.response_timeout)
        uid = kwargs.pop("rdisq_uid", None)
        method_queue_name = self.service_class.get_queue_name_for_method(method_name, uid)
        redis_con = self.service_class.redis_dispatcher.get_redis()
        task_id = method_queue_name + generate_task_id()
        request_payload = RequestPayload(
            task_id=task_id,
            args=args,
            kwargs=kwargs,
            timeout=timeout
        )
        request_key = get_request_key(task_id)
        serialized_request = self.service_class.serializer.dumps(request_payload)
        redis_con.setex(request_key, timeout, serialized_request)
        redis_con.lpush(method_queue_name, task_id)
        return RdisqResponse(task_id, self)

    def get_stub_method(self, method_name):
        raise NotImplementedError()


class RdisqAsyncConsumer(AbstractRdisqConsumer):
    def get_stub_method(self, method_name):
        def c(*args, **kwargs):
            return self.send(method_name, *args, **kwargs)
        return c


class RdisqWaitingConsumer(AbstractRdisqConsumer):
    def get_stub_method(self, method_name):
        def c(*args, **kwargs):
            return self.send(method_name, *args, **kwargs).wait()
        return c



__author__ = 'smackware'
from unittest.case import TestCase
from rdisq.response import RdisqResponse
from rdisq import RESULT_ATTR
from rdisq import PROCESS_TIME_ATTR
from rdisq import EXCEPTION_ATTR

class RdisqTest(TestCase):
    def test__result__process_response(self):
        valid_response = {
            RESULT_ATTR: "RETURNED",
            PROCESS_TIME_ATTR: 2,
            EXCEPTION_ATTR: None,
        }
        result = RdisqResponse(1, None)
        returned_value = result.process_response_data(valid_response)
        self.assertEqual(result.exception, None)
        self.assertEqual(returned_value, "RETURNED")

    def test__result__process_response_exception(self):
        class TestException(Exception):
            pass

        valid_response = {
            RESULT_ATTR: None,
            PROCESS_TIME_ATTR: 2,
            EXCEPTION_ATTR: TestException("BLAH"),
        }
        result = RdisqResponse(1, None)
        response = result.process_response_data(valid_response)
        self.assertEqual(response, None)
        self.assertTrue(isinstance(result.exception, TestException))



__author__ = 'smackware'

import os
import uuid


def get_mac():
    return uuid.getnode()


def get_consumer_id():
    return "%s-%s" % (get_mac(), os.getpid(), )


def get_request_key(task_id):
    return "request_%s" % (task_id, )


def generate_task_id():
    return "%s-%s" % (get_consumer_id(), uuid.uuid4().hex, )





__author__ = 'smackware'

import time
import uuid


MISSING_DISPATCHER_ERROR_TEXT = \
    "Service class must have a 'redis_dispatcher' attribute pointing to a RedisDispatcher instance"
MISSING_SERVICE_NAME_IN_MAIN_ERROR_TEXT = \
    "When instantiating a service in __main__, the service_name class attribute must be set"

from . import EXPORTED_METHOD_PREFIX

from .payload import RequestPayload
from .payload import ResponsePayload

from .identification import get_request_key
from .serialization import PickleSerializer

from .redis_dispatcher import LocalRedisDispatcher
from .consumer import RdisqAsyncConsumer
from .consumer import RdisqWaitingConsumer


# Decorator
def remote_method(callable_object):
    callable_object.is_remote = True
    return callable_object


class RdisqService(object):
    """
    Inheritance example:
    =============================================
    class CalculatorService(RdisqService):
        service_name = "CalculatorService"
        redis_dispatcher = LocalRedisDispatcher()

        @remote_method
        def add(self, a, b):
            return a + b;
    =============================================
    """
    logger = None
    log_returned_exceptions = True
    service_name = None
    response_timeout = 10
    redis_dispatcher = None
    serializer = PickleSerializer()
    __go = True
    __sync_consumer = None
    __async_consumer = None

    def __init__(self, uid=None):
        if self.redis_dispatcher is None:
            raise NotImplementedError(MISSING_DISPATCHER_ERROR_TEXT)
        if self.__class__.__module__ == '__main__' and self.service_name is None:
            raise NotImplementedError(MISSING_SERVICE_NAME_IN_MAIN_ERROR_TEXT)
        self.__uid = uid or str(uuid.uuid4())
        self.__queue_to_callable = None
        self.__map_exposed_methods_to_queues()

    @property
    def rdisq_go(self):
        return self.__go

    def rdisq_process_one(self):
        self.__process_one()


    @classmethod
    def __set_service_name(cls, service_name):
        cls.service_name = service_name

    @classmethod
    def get_service_name(cls):
        if cls.service_name is None:
            cls.service_name = "%s.%s" % (cls.__module__, cls.__name__)
        return cls.service_name

    def __map_exposed_methods_to_queues(self):
        self.__queue_to_callable = self.get_queue_name_to_exposed_method_mapping()
        if not self.__queue_to_callable:
            raise AttributeError("Cannot instantiate a service with no exposed methods")

    @classmethod
    def get_method_if_exists(cls, name):
        method = cls.__dict__.get(name)
        if method is None or not hasattr(method, '__call__'):
            return None
        return method

    @staticmethod
    def is_remote_method(method):
        if method is None or not hasattr(method, '__call__'):
            return False
        if hasattr(method, 'is_remote'):
            return True
        # Legacy compatibility, remote methods used to be prefixed with 'q_'
        if method.__name__.startswith(EXPORTED_METHOD_PREFIX):
            return True
        return False

    def get_queue_name_to_exposed_method_mapping(self):
        mapping = {}
        for attr in dir(self):
            call = getattr(self, attr)
            if not self.is_remote_method(call):
                continue
            stub_method_name = self.chop_prefix_from_exported_method_name(attr)
            method_queue_name = self.get_queue_name_for_method(stub_method_name)
            mapping[method_queue_name] = call
            method_queue_name = self.get_queue_name_for_method(stub_method_name, self.__uid)
            mapping[method_queue_name] = call
        return mapping

    @classmethod
    def get_consumer(cls):
        if cls.__sync_consumer is None:
            cls.__sync_consumer = RdisqWaitingConsumer(cls)
        return cls.__sync_consumer

    @classmethod
    def get_async_consumer(cls):
        if cls.__async_consumer is None:
            cls.__async_consumer = RdisqAsyncConsumer(cls)
        return cls.__async_consumer

    @classmethod
    def get_redis(cls):
        return cls.redis_dispatcher.get_redis()


    @classmethod
    def get_queue_name_for_method(cls, method_name, prefix=None):
        if prefix is not None:
            return prefix + "_" +cls.get_service_name() + "_" + method_name
        return cls.get_service_name() + "_" + method_name
        
    @classmethod
    def chop_prefix_from_exported_method_name(cls, method_name):
        # Legacy compatibility, remote methods used to be prefixed with 'q_'
        if method_name.startswith(EXPORTED_METHOD_PREFIX):
            return method_name[len(EXPORTED_METHOD_PREFIX):]
        return method_name

    def pre(self, method_queue_name):
        """Performs after something was found in the queue"""
        pass

    def post(self, method_queue_name):
        """Performs after a queue fetch and process"""
        pass

    def on_exception(self, exc):
        pass

    def on_start(self):
        pass

    def __process_one(self, timeout=0):
        """Process a single queue event
        Will pend for an event (unless timeout is specified) then it will process it
        """
        redis_con = self.get_redis()
        redis_result = redis_con.brpop(self.__queue_to_callable.keys(), timeout=timeout)
        if redis_result is None:  # Timeout
            return
        method_queue_name, task_id = redis_result
        request_key = get_request_key(task_id.decode())
        call = self.__queue_to_callable[method_queue_name.decode()]
        data_string = redis_con.get(request_key)
        if data_string is None:
            return
        self.pre(method_queue_name)
        request_payload = self.serializer.loads(data_string)
        timeout = request_payload.timeout
        if request_payload.task_id != task_id.decode():
            raise ValueError("Memorized task id is mismatching to received-payload task_id")
        args = request_payload.args
        kwargs = request_payload.kwargs
        time_start = time.time()
        try:
            result = call(*args, **kwargs)
            raised_exception = None
        except Exception as ex:
            result = None
            raised_exception = ex
            if self.log_returned_exceptions and self.logger:
                self.logger.exception(ex)
            self.on_exception(ex)
        duration_seconds = time.time() - time_start
        response_payload = ResponsePayload(
            returned_value=result,
            processing_time_seconds=duration_seconds,
            raised_exception=raised_exception
        )
        serialized_response = self.serializer.dumps(response_payload)
        redis_con.lpush(task_id, serialized_response)
        redis_con.expire(task_id, timeout)
        self.post(method_queue_name)

    @classmethod
    def get_service_uid_list_key(cls):
        return "rdisq_uids:" + cls.get_service_name()

    @classmethod
    def list_uids(cls):
        uids = []
        rdb = cls.get_redis()
        key = cls.get_service_uid_list_key()
        for k,v in rdb.hgetall(key).items():
            if float(v) > time.time() - 10:
                uids.append(k.decode())
            else:
                rdb.hdel(key, k)
        return uids

    def process(self):
        self.on_start()
        redis_con = self.get_redis()
        while self.__go:
            self.__process_one(1)
            redis_con.hset(self.get_service_uid_list_key(), self.__uid, time.time())

    def stop(self):
        self.__go = False



__author__ = 'smackware'

from redis import Redis
from redis import ConnectionPool


class AbstractRedisDispatcher(object):

    def __init__(self, *args, **kwargs):
        pass

    def get_redis(self, *args, **kwargs):
        """
        Produce an instance of an active redis connection
        """
        raise NotImplementedError("Must implement get_redis(self) method of Rdisq subclass")


class SimpleRedisDispatcher(AbstractRedisDispatcher):
    redis_args = None
    redis_kwargs = None

    def __init__(self, *redis_args, **redis_kwargs):
        self.redis_args = redis_args
        self.redis_kwargs = redis_kwargs
        AbstractRedisDispatcher.__init__(self)

    def get_redis(self):
        return Redis(*self.redis_args, **self.redis_kwargs)


class LocalRedisDispatcher(SimpleRedisDispatcher):
    def __init__(self):
        SimpleRedisDispatcher.__init__(self)


class PoolRedisDispatcher(AbstractRedisDispatcher):
    redis_pool = None

    def __init__(self, *pool_args, **pool_kwargs):
        self.redis_pool = ConnectionPool(*pool_args, **pool_kwargs)
        AbstractRedisDispatcher.__init__(self)

    def get_redis(self):
        return Redis(connection_pool=self.redis_pool)



__author__ = 'smackware'

import time


class RdisqResponseTimeout(Exception):
    task_id = None

    def __init__(self, task_id):
        self.task_id = task_id


class RdisqResponse(object):
    _task_id = None
    rdisq_consumer = None
    response_payload = None
    total_time_seconds = None
    called_at_unixtime = None
    timeout = None
    exception = None
    returned_value = None

    def __init__(self, task_id, rdisq_consumer):
        self._task_id = task_id
        self.rdisq_consumer = rdisq_consumer
        self.called_at_unixtime = time.time()

    def get_service_timeout(self):
        return self.rdisq_consumer.service_class.response_timeout

    def is_processed(self):
        if self.response_payload is not None:
            return True
        redis_con = self.rdisq_consumer.service_class.redis_dispatcher.get_redis()
        return redis_con.llen(self._task_id) > 0

    def is_exception(self):
        return self.response_payload.raised_exception is not None

    @property
    def process_time_seconds(self):
        return self.response_payload.processing_time_seconds

    @property
    def exception(self):
        return self.response_payload.raised_exception

    def wait(self):
        timeout = self.get_service_timeout()
        redis_con = self.rdisq_consumer.service_class.redis_dispatcher.get_redis()
        redis_response = redis_con.brpop(self._task_id, timeout=timeout)  # can be tuple of (queue_name, string) or None
        if redis_response is None:
            raise RdisqResponseTimeout(self._task_id)
        queue_name, response = redis_response
        self.total_time_seconds = time.time() - self.called_at_unixtime
        response_payload = self.rdisq_consumer.service_class.serializer.loads(response)
        redis_con.delete(self._task_id)
        self.response_payload = response_payload
        if self.is_exception():
            raise self.exception
        return self.response_payload.returned_value


__author__ = 'smackware'

from collections import namedtuple

RequestPayload = namedtuple("RequestPayload", "task_id timeout args kwargs")
ResponsePayload = namedtuple("ResponsePayload", "returned_value raised_exception processing_time_seconds")


import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "rdisq",
    version = "0.0.8",
    author = "Lital Natan",
    author_email = "litaln@gmail.com",
    description = ("Super minimal workload distribution framework over redis queues"),
    license = "MIT",
    install_requires = ["redis>=3.3.11"],
    keywords = "",
    packages=find_packages(),
    long_description="Please see README.md",
    url="https://github.com/smackware/rdisq",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
